/*
    Conditional edge creation utility for Ldraw
    by Lee Gaiteri (LummoxJR@aol.com)
    http://www.lummoxjr.com

    Version 0.7: March 2001

    Usage:  java edger input_file.dat >output_file.dat

    Edges along only to one shape are given a solid line, but in white; these
    shouldn't exist in a normal part, so they're output in white (color 15).
    Normally these indicate a problem in the part.

    Edges that exist along two or more shapes are given a conditional line.
    An edge should be shared by two and only two shapes. Conditional lines
    are output as color 24.

    - Version 0.7: Conditional edges are not output if their reference points
                   are on the same plane with the line.
    - Version 0.6: Bug fix
    - Version 0.5: Initial release. Checks for overlapping edges added.
 */

import java.io.*;
import java.util.*;

public class edger {
    Vector edges;
    Vector done;

    public static void main(String args[]) {
        if(args.length==0) return;
        String f=loadFile(args[0]);
        if(f==null) return;
        edger obj=new edger();
        StringTokenizer stok=new StringTokenizer(obj.generate(f),"\n");
        while(stok.hasMoreTokens()) System.out.println(stok.nextToken());
        }

    public edger() {
        edges=new Vector();
        done=new Vector();
        }

    public static String loadFile(String path,String filename) {
        return loadFile((new File(path,filename)).getAbsolutePath());
        }
    public static String loadFile(String filename) {
        try {
            int i;
            String item,line;
            StringBuffer docbuffer=new StringBuffer();
            FileInputStream fis=new FileInputStream(filename);
            BufferedReader reader=new BufferedReader(new InputStreamReader(fis));
            while((line=reader.readLine())!=null) {
                docbuffer.append(line.trim());
                docbuffer.append("\n");
                }
            reader.close();
            fis.close();
            return docbuffer.toString();
            }
        catch(Exception e) {
            System.out.println(e.toString());
            return "";
            }
        }

    public int findEdge(String v1,String v2) {
        int i,len=edges.size();
        String v[];
        for(i=0;i<len;++i) {
            v=(String[])edges.elementAt(i);
            if(v[0].equals(v1) && v[1].equals(v2)) return i;
            if(v[1].equals(v1) && v[0].equals(v2)) return i;
            }
        return -1;
        }

    public int findFinishedEdge(String v1,String v2) {
        int i,len=done.size();
        String v[];
        for(i=0;i<len;++i) {
            v=(String[])done.elementAt(i);
            if(v[0].equals(v1) && v[1].equals(v2)) return i;
            if(v[1].equals(v1) && v[0].equals(v2)) return i;
            }
        return -1;
        }

	// add an edge point to the list, plus a third point representing a connecting shape
	public void addedge(String v1,String v2,String v3) {
		String e[];
		int k=findFinishedEdge(v1,v2);
		if(k>=0) return;
		k=findEdge(v1,v2);
		if(k<0) {
			e=new String[4]; e[0]=v1; e[1]=v2; e[2]=v3; e[3]=null;
			edges.addElement(e); return;
			}
		e=(String[])edges.elementAt(k);
		edges.removeElementAt(k);
		e[3]=v3;
		done.addElement(e);
		}

	// search through the list for overlapping lines
	public void checkoverlaps() {
		int i,j;
		String o[];
		for(i=0;i<edges.size()-1;++i) {
			for(j=i+1;j<edges.size();++j) {
				o=overlap((String[])edges.elementAt(i),(String[])edges.elementAt(j));
				if(o!=null) {splitedges(o); --i; break;}
				}
			}
		}

    public void checkplanes() {
		int i;
		String o[];
		for(i=0;i<done.size()-1;++i) {
    		o=(String[])done.elementAt(i);
            if(o.length!=4) continue;
            if(coplanar(o[0],o[1],o[2],o[3])) done.removeElementAt(i--);
			}
        }

	// return the points in the order that they occur, or null if they don't overlap
	public static String[] overlap(String e1[],String e2[]) {
		if(e1[0].equals(e1[1]) || e2[0].equals(e2[1])) return null;    // doesn't count as an edge
		if(!colinear(e1[0],e1[1],e2[0])) return null;
		if(!colinear(e1[0],e1[1],e2[1])) return null;
		String se[]=new String[4],swap;
		se[0]=e1[0]; se[1]=e1[1]; se[2]=e2[0]; se[3]=e2[1];
		// simple bubble sort; it's only 4 values
		if(compareVertex(se[0],se[1])>0) {swap=se[0]; se[0]=se[1]; se[1]=swap;}
		if(compareVertex(se[1],se[2])>0) {swap=se[2]; se[2]=se[1]; se[1]=swap;}
		if(compareVertex(se[2],se[3])>0) {swap=se[2]; se[2]=se[3]; se[3]=swap;}
		if(compareVertex(se[0],se[1])>0) {swap=se[0]; se[0]=se[1]; se[1]=swap;}
		if(compareVertex(se[1],se[2])>0) {swap=se[2]; se[2]=se[1]; se[1]=swap;}
		if(compareVertex(se[0],se[1])>0) {swap=se[0]; se[0]=se[1]; se[1]=swap;}
		// check for separate lines; they may touch but if they don't overlap, return null
		if(se[0].equals(e1[0]) && se[1].equals(e1[1])) return null;
		if(se[0].equals(e1[1]) && se[1].equals(e1[0])) return null;
		if(se[0].equals(e2[0]) && se[1].equals(e2[1])) return null;
		if(se[0].equals(e2[1]) && se[1].equals(e2[0])) return null;
		return se;
		}

	/*
		The overlap() function returns 4 sorted colinear points, at least 3
		of which should be unique. Edges may exist at any interval except
		where two end points are identical. The longest possible edge will be
		ol[0]->ol[3]. Working inwards from the longest edge, take three
		unique points in order and find an existing edge (completed or matched
		to just one shape) along the outer two points; split it into two
		separate edges at the intermediate point.
		The nested loop is constructed such that whenever a line is split, if
		a second split is required it will be done in a later loop pass. At
		most we will only split a line into three parts total.
	 */
	public void splitedges(String ol[]) {
		String theedge[],newedge[];
		int i,j,k,e1,e2;
		for(j=3;j>=2;--j) {
			if(ol[j].equals(ol[j-1])) continue;
			for(i=0;j-i>=2;++i) {
				if(ol[i].equals(ol[i+1])) continue;
				if((k=findEdge(ol[i],ol[j]))>=0) {
					theedge=(String[])edges.elementAt(k);
					edges.removeElementAt(k);
					addedge(theedge[0],ol[j-1],theedge[2]);
					addedge(theedge[1],ol[j-1],theedge[2]);
					}
				if((k=findFinishedEdge(ol[i],ol[j]))>=0) {
					theedge=(String[])done.elementAt(k);
					newedge=new String[theedge.length];
					System.arraycopy(theedge,1,newedge,1,theedge.length-1);
					theedge[1]=newedge[0]=ol[j];
					done.addElement(newedge);
					}
				}
			}
		}

	public static boolean colinear(String v1,String v2,String v3) {
		int i;
		double points[],ratio,small=1.0e-8;
		StringTokenizer stok;
		stok=new StringTokenizer(v1+" "+v2+" "+v3," ");
		points=new double[9];
		for(i=0;i<9;++i) {
			points[i]=(new Double(stok.nextToken())).doubleValue();
			}
		// if any two vertices are equal, then the three are colinear
		if(points[0]==points[3] && points[1]==points[4] && points[2]==points[5]) return true;
		if(points[0]==points[6] && points[1]==points[7] && points[2]==points[8]) return true;
		if(points[6]==points[3] && points[7]==points[4] && points[8]==points[5]) return true;
		if(points[0]==points[3]) {
			if(points[0]!=points[6]) return false;  // if x1==x2, x1==x2==x3 or else not colinear
			// if y1==y2, y1==y2==y3 or else not colinear
			// if x1==x2==x3 and y1==y2==y3, the points are colinear
			if(points[1]==points[4]) return (points[1]==points[7]);
			ratio=(points[7]-points[1])/(points[4]-points[1]);
			if(Math.abs(ratio*(points[5]-points[2])-(points[8]-points[2]))>small) return false;
			return true;
			}
		ratio=(points[6]-points[3])/(points[3]-points[0]);
		if(Math.abs(ratio*(points[4]-points[1])-(points[7]-points[1]))>small) return false;
		if(Math.abs(ratio*(points[5]-points[2])-(points[8]-points[2]))>small) return false;
		return true;
		}

    public static boolean coplanar(String v1,String v2,String v3,String v4) {
		int i,j;
		double points[],dir[],small=1.0e-8;
		StringTokenizer stok;
		stok=new StringTokenizer(v1+" "+v2+" "+v3+" "+v4," ");
		points=new double[12];
        dir=new double[3];
        for(i=0;i<12;++i) points[i]=(new Double(stok.nextToken())).doubleValue();
		// if any two vertices are equal, then the four are coplanar
		for(i=0;i<9;i+=3) for(j=i+3;j<12;j+=3)
            if(points[i]==points[j] && points[i+1]==points[j+1] && points[i+2]==points[j+2]) return true;
        // adjust points to make the math easier
        for(i=3;i<12;++i) points[i]-=points[i%3];
        dir[0]=points[4]*points[8]-points[5]*points[7];
        dir[1]=points[5]*points[6]-points[3]*points[8];
        dir[2]=points[3]*points[7]-points[4]*points[6];
        return (Math.abs( dir[0]*points[9]+dir[1]*points[10]+dir[2]*points[11] )<small);
        }

	// <0 if v1<v2, =0 if v1=v2, >0 if v1>v2
	public static int compareVertex(String v1,String v2) {
		double d1,d2;
		StringTokenizer stok1,stok2;
		stok1=new StringTokenizer(v1," ");
		stok2=new StringTokenizer(v2," ");
		d1=(new Double(stok1.nextToken())).doubleValue();
		d2=(new Double(stok2.nextToken())).doubleValue();
		if(d1<d2) return -1; if(d1>d2) return 1;
		d1=(new Double(stok1.nextToken())).doubleValue();
		d2=(new Double(stok2.nextToken())).doubleValue();
		if(d1<d2) return -1; if(d1>d2) return 1;
		d1=(new Double(stok1.nextToken())).doubleValue();
		d2=(new Double(stok2.nextToken())).doubleValue();
		if(d1<d2) return -1; if(d1>d2) return 1;
		return 0;
		}

    public String generate(String file) {
        StringBuffer buf=new StringBuffer();
        String line,type,n1,n2,n3,ngroup,vertices[],o[];
        edges=new Vector();
        done=new Vector();
        int i,j,k;
        StringTokenizer stok=new StringTokenizer(file,"\n"),stok2;
        while(stok.hasMoreTokens()) {
            line=stok.nextToken().trim();
            stok2=new StringTokenizer(line," \t");
            type=stok2.nextToken(); if(type==null) continue;
            if(type.equals("0") || type.equals("1")) continue;
            if(type.equals("2")) j=2;
            else if(type.equals("3")) j=3;
            else if(type.equals("4") || type.equals("5")) j=4;
            else continue;
            if(stok2.nextToken()==null) continue;
            vertices=new String[j];
            for(i=0;i<j;++i) {
                n1=stok2.nextToken(); if(n1==null) break;
                n2=stok2.nextToken(); if(n2==null) break;
                n3=stok2.nextToken(); if(n3==null) break;
                vertices[i]=n1+" "+n2+" "+n3;
                }
            if(i<j) {vertices=null; continue;}
            if(j==2 || type.equals("5")) {
                // if we've already got an edge line ready to go for this,
                // chuck it in favor of the new one
                k=findFinishedEdge(vertices[0],vertices[1]);
                if(k>=0) done.removeElementAt(k);
                // remove any edges in progress, too
                k=findEdge(vertices[0],vertices[1]);
                if(k>=0) edges.removeElementAt(k);
                // since this array will be length 3 or 5, it won't be output as a conditional
                o=new String[j+1]; System.arraycopy(vertices,0,o,0,j);
                o[j]=null;
                done.addElement(o);
                vertices=null; continue;
                }
            for(i=0;i<j;++i) addedge(vertices[i],vertices[(i+1)%j],vertices[(i+2)%j]);
            vertices=null;
            }
        checkoverlaps();
        checkplanes();
        Enumeration e;
        if(edges.size()>0) buf.append("0 WRITE Unmatched edges\n");
        e=edges.elements();
        while(e.hasMoreElements()) {
            o=(String[])e.nextElement();
            buf.append("2 15 ").append(o[0]).append(" ").append(o[1]).append("\n");
            }
        if(done.size()>0) buf.append("0 WRITE Conditional edges\n");
        e=done.elements();
        while(e.hasMoreElements()) {
            o=(String[])e.nextElement();
            if(o.length!=4) continue;   // don't repeat a line from the original file
            // output all conditional lines
            buf.append("5 24 ").append(o[0]).append(" ").append(o[1]).append(" ").append(o[2]).append(" ").append(o[3]);
            buf.append("\n");
            }
        return buf.toString();
        }
}